#include <Windows.h>
#include "toolfe.h"

using namespace win32fe;

tool::~tool(void) {
  if (pathdll) {
    FreeLibrary(pathdll);
  }
}

tool::tool(void) {
  tool::OptionTags= "hnpuvwN";
  tool::Options['h'] = &tool::FoundHelp;
  tool::Options['n'] = &tool::FoundNoDetect;
  tool::Options['p'] = &tool::FoundPath;
  tool::Options['u'] = &tool::FoundUse;
  tool::Options['v'] = &tool::FoundV;
  tool::Options['w'] = &tool::FoundWoff;
  tool::Options['N'] = &tool::FoundNT4;
  tool::Options[UNKNOWN] = &tool::FoundUnknown;
  version_string  = "Win32 Development Tool Front End, version 1.2.4.";
  detect       = true;
  woff         = false;
  verbose      = false;
  helpfound    = false;
  versionfound = false;
  inpath       = false;
  NT4          = false;
}
  
void tool::GetArgs(int argc,char *argv[]) {
  for (int i=2;i<argc;i++) {
    std::string temp=argv[i];
    if (temp=="--verbose") {
      verbose=true;
      std::cout << std::endl;
    }
    arg.push_back(argv[i]);
  }
  if (argc == 2) {
    if (!((arg.front()=="--version") || (arg.front()=="--help"))) {
      arg.push_back("--help");
    }
  }
}

int tool::Parse() {
  LI i = arg.begin();
  std::string thetool = *i;
  arg.pop_front();
  pathdll = LoadLibrary("win32fepath.dll");
  if (pathdll) {
    getshortpath = GetProcAddress(pathdll,"GetShortPath");
    getlongpath  = GetProcAddress(pathdll,"GetLongPath");
    if ((!getshortpath) || (!getlongpath)) {
      arg.push_front("--NT4");
    }
  } else {
    std::cerr << "win32fe Error: Unable to find pathdll.dll." << std::endl;
    std::cerr << "        It should be in the same directory as win32fe.exe." << std::endl;
    return(-1);
  }
  arg.push_front(thetool);
  i = arg.begin();
  while (i!=arg.end()) {
    std::string temp = *i;
    if (temp.substr(0,2)=="--") {
      char flag = temp[2];
      if (tool::OptionTags.find(flag)==std::string::npos) {
        (this->*tool::Options[UNKNOWN])(i);
      } else {
        (this->*tool::Options[flag])(i);
      }
    } else {
      i++;
    }
  }
  if (IsAKnownTool()) {
    AddSystemInfo();
  } else {
    if (!(helpfound || versionfound)) {
      if (arg.begin()!=arg.end()) {
        std::cerr << std::endl << "Error: win32fe: Unknown Tool: " << arg.front() << std::endl;
        std::cerr << "  Use --help for more information on win32fe options." << std::endl << std::endl;
      } else {
        std::cerr << std::endl << "Error: win32fe: Unknown Tool:" << std::endl;
        std::cerr << "  Use --help for more information on win32fe options." << std::endl << std::endl;
      }
    }
  }
  return(0);
}

void tool::AddSystemInfo(void) {
  if (detect) {
    FindInstallation();
    AddPaths();
  }
}

void tool::FindInstallation(void) {
  /* Find location of tool */
  InstallDir = arg.front();
  bool ierr = false;
  
    /* First check if a full path was specified with --use */
  std::string::size_type n = InstallDir.find(":");
  if (n==std::string::npos) {
    std::string extension = ".exe";
    char tmppath[MAX_PATH],*tmp; /* Windows.h */ 
    if (SearchPath(NULL,InstallDir.c_str(),extension.c_str(),MAX_PATH,tmppath,&tmp)) { /* Windows.h */
      /*
        Note: There are 6 locations searched, the 6th of these is the PATH variable.
        But, in reality, the first 5 don't make any sense for this tool, so it is 
        assumed that the tool was found in the PATH variable.
      */
      //InstallDir = (std::string)tmppath;
	InstallDir = tmppath;
	} else {
      ierr = true;
    }
  }
  std::string tool=arg.front();
  n = tool.rfind("\\")+1;
  tool = tool.substr(n,tool.rfind(".")-n);

  if (GetShortPath(InstallDir)) {
    inpath = true;
    n = InstallDir.rfind("\\BIN\\");
    if (n==std::string::npos) {
      n = InstallDir.rfind("\\Bin\\");
    }
    if (n==std::string::npos) {
      n = InstallDir.rfind("\\bin\\");
    }
    InstallDir = InstallDir.substr(0,n+1);
    if (verbose) {
      std::string longpath;
      GetLongPath(InstallDir,longpath);
      std::cout << "win32fe: " << tool << " Installation Found: " << longpath << std::endl;
    }
  } else {
    ierr = true;
  }
  if (ierr) {
    if (!woff) {
      std::cerr << std::endl << "Warning: win32fe: Tool Not Found with the PATH environment variable: " << tool << std::endl;
      std::cerr << "  Specify the complete path to " << tool << " with --use or --path" << std::endl;
      std::cerr << "  Use --help for more information on win32fe options." << std::endl << std::endl;
    }
  }
}

void tool::AddPaths(void) {
  if (!inpath) {
    arg.push_back("--path");
    LI i = arg.end();
    i--;
    arg.push_back(InstallDir);
    FoundPath(i);
  }
}

int tool::Execute(void) {
  if (helpfound) {
    Help();
    return(0);
  }
  if (versionfound || verbose) {
    DisplayVersion();
  }
  return(0);
}

void tool::Help(void) {
  std::cout << std::endl << version_string << std::endl << std::endl;  
  std::cout << "Usage: win32fe <tool> --<win32fe options> -<tool options> <files>" << std::endl;
  std::cout << "  <tool> must be the first argument to win32fe" << std::endl << std::endl;
  std::cout << "<tool>: {cl,icl,df,f90,ifl,bcc32,lib,tlib}" << std::endl;
  std::cout << "  cl:    Microsoft 32-bit C/C++ Optimizing Compiler" << std::endl;
  std::cout << "  icl:   Intel C/C++ Optimizing Compiler" << std::endl;
  std::cout << "  df:    Compaq Visual Fortran Optimizing Compiler" << std::endl;
  std::cout << "  f90:   Compaq Visual Fortran90 Optimizing Compiler" << std::endl;
  std::cout << "  ifl:   Intel Fortran Optimizing Compiler" << std::endl;
  std::cout << "  bcc32: Borland C++ for Win32" << std::endl;
  std::cout << "  lib:   Microsoft Library Manager" << std::endl;
  std::cout << "  tlib:  Borland Library Manager" << std::endl << std::endl;
  std::cout << "<win32fe options>: {help,nodetect,path,use,verbose,version,woff}" << std::endl;
  std::cout << "  --help:       Output this help message and help for <tool>" << std::endl;
  std::cout << "  --nodetect:   Suppress automatic detection of <tool> installation" << std::endl;
  std::cout << "  --path <arg>: <arg> specifies an addition to the PATH that is required" << std::endl;
  std::cout << "                (ex. the location of a required .dll)" << std::endl;
  std::cout << "  --use <arg>:  <arg> specifies the variant of <tool> to use" << std::endl;
  std::cout << "  --verbose:    Echo to stdout the translated commandline" << std::endl;
  std::cout << "  --version:    Output version info for win32fe and <tool>" << std::endl;
  std::cout << "  --woff:       Suppress win32fe specific warning messages" << std::endl;
  std::cout << "  --NT4:        Enable usage of win32fe on WindowsNT4 machines" << std::endl << std::endl;
  std::cout << "=========================================================================" << std::endl << std::endl;
}

void tool::FoundUnknown(LI &i) {
  i++;
}

void tool::FoundPath(LI &i) {
  if (*i=="--path") {
    i = arg.erase(i);
    if (i!=arg.end()) {
      /* Note 1024 might not be sufficiently large */
      int length = 1024;
      char buff[1024];
      std::string path="PATH";
      GetEnvironmentVariable(path.c_str(),buff,length); /* Windows.h */
      std::string newpath = (std::string)buff;

      std::string shortpath,inpath = *i;
      std::string::size_type n=inpath.find(";");
      std::string::size_type nold=0;
      do {
        std::string longpath,pathi=inpath.substr(nold,n);
        if (GetShortPath(pathi)) {
          if (verbose) {
            GetLongPath(pathi,longpath);
            std::cout << "win32fe: Adding to path: " << longpath << std::endl;
          }
          shortpath += pathi + ";";
        } else {
          if (!woff) {
            std::cout << "Warning: win32fe Path Not Found: " << inpath.substr(nold,n) << std::endl;
          }
        }
        nold = n;
        n = inpath.find(";",nold+1);
      } while (nold++!=std::string::npos);
      newpath = shortpath + newpath;
      SetEnvironmentVariable(path.c_str(),newpath.c_str()); /* Windows.h */
      i = arg.erase(i);
    } else {
      i--;
      arg.push_back("--help");
      i--;
    }
  }
}

void tool::FoundNoDetect(LI &i) {
  if (*i=="--nodetect") {
    detect = false;
    i = arg.erase(i);
    if (verbose) {
      std::cout << "win32fe: Turing off installation detection" << std::endl;
    }
  } else {
    i++;
  }
}

void tool::FoundUse(LI &i) {
  if (*i=="--use") {
    i = arg.erase(i);
    if (i!=arg.end()) {
      ReplaceSlashWithBackslash(*i);
      arg.pop_front();
      arg.push_front(*i);
      i = arg.erase(i);
    } else {
      i--;
      arg.push_back("--help");
      i--;
    }
  }
}

void tool::FoundWoff(LI &i) {
  if (*i == "--woff") {
    woff = true;
    i = arg.erase(i);
  } else {
    i++;
  }
}

void tool::FoundNT4(LI &i) {
  if (*i == "--NT4") {
    FARPROC gsp,glp;
    gsp = GetProcAddress(pathdll,"GetShortPath_nt4");
    glp  = GetProcAddress(pathdll,"GetLongPath_nt4");
    if (gsp && glp) {
      getshortpath = gsp;
      getlongpath  = glp;
      if (verbose) {
         std::cout << "Enabling use of NT4 versions of path manipulation routines." << std::endl;
      }
    } else {
      std::cerr << "Warning: win32fe unable to retrieve NT4 path manipulation routines." << std::endl;
      std::cerr << "         Problem with win32fepath.dll" << std::endl << std::endl;
    }
    i = arg.erase(i);
  } else {
    i++;
  }
}
void tool::FoundV(LI &i) {
  std::string temp = *i;
  if (temp == "--verbose") {
    verbose=true;
    i = arg.erase(i);
  } else if (temp == "--version") {
    versionfound=true;
    i = arg.erase(i);
  } else {
    i++;
  }
}

void tool::FoundHelp(LI &i) {
  if (*i == "--help") {
    helpfound = true;
    i = arg.erase(i);
  } else {
    i++;
  }
}

void tool::FoundFile(LI &i) {
  std::string temp = *i;
  ReplaceSlashWithBackslash(temp);
  /* If file specified with <path>\<filename> then verify it exists */
  std::string::size_type n = temp.rfind("\\");
  if (n!=std::string::npos) {
    std::string dir = temp.substr(0,n);
    if (GetShortPath(dir)) {
      temp = dir + temp.substr(n);
    } else {
      std::cerr << "Warning: win32fe: File Not Found: " << temp << std::endl;
    }
  }
  file.push_back(temp);
}

void tool::ReplaceSlashWithBackslash(std::string &name) {
//   for (std::string::size_type i=0;i<name.length();i++)
//     if (name[i]=='/') name[i]='\\';
}

void tool::ProtectQuotes(std::string &name) {
  std::string::size_type a,b;
  a = name.find("\"");
  if (a!=std::string::npos) {
    std::string temp = name.substr(0,a+1);
    temp += "\\\"";
    temp += name.substr(a+1,std::string::npos);
    name = temp;
    b = name.rfind("\"");
    if (b!=a+2) {
      temp = name.substr(0,b);
      temp += "\\\"";
      temp += name.substr(b,std::string::npos);
      name = temp;
    }
  }
}

void tool::RemoveQuotes(std::string &name) {
  if ((name[0]=='\'') || (name[0]=='\"')) {
    name=name.substr(1);
  }
  std::string::size_type n=name.length();
  if ((name[n-1]=='\'') || (name[n-1]=='\"')) {
    name=name.substr(0,n-1);
  }
}

int tool::GetShortPath(std::string &name) {
  int (*gsp)(const char*,const int,char *);
  char shortpath[MAX_PATH]; /* Windows.h */

  gsp = reinterpret_cast<int (*)(const char*,const int,char*)>(getshortpath);

  int size = gsp(name.c_str(),MAX_PATH,shortpath); /* Windows.h */
  if ((size>0)&&(size<=MAX_PATH)) {                 /* Windows.h */
    name =(std::string)shortpath;
  }
  return(size);
}

int tool::GetLongPath(std::string &shortpath,std::string &longpath) {
  int (*glp)(const char*,const int,char *);
  char longpath_c[MAX_PATH]; /* Windows.h */

  glp = reinterpret_cast<int (*)(const char *,const int,char *)>(getlongpath);
  int size = glp(shortpath.c_str(),MAX_PATH,longpath_c); /* Windows.h */
  if ((size>0) && (size<=MAX_PATH)) {                    /* Windows.h */
    longpath = (std::string)longpath_c;
  }
  return(size);
}

void tool::PrintListString(std::list<std::string> &liststr) {
  std::cout << "Printing..." << std::endl;
  LI i = liststr.begin();
  while (i!=liststr.end()) std::cout << *i++ << " ";
  std::cout << std::endl;
}

void tool::Merge(std::string &str,std::list<std::string> &liststr,LI &i) {
  while (i!=liststr.end()) {
    str += " " + *i++;
  }
}

void tool::DisplayVersion(void) {
  std::cout << version_string << std::endl;
}

bool tool::IsAKnownTool(void) {
  return(false);
}
